importPackage(java.lang);
importPackage(java.sql);
importPackage(java.util);
importPackage(java.io);
importPackage(java.net);
importPackage(Packages.javax.xml);
importPackage(Packages.javax.xml.parsers);
importPackage(Packages.javax.xml.transform);
importPackage(Packages.javax.xml.transform.dom);
importPackage(Packages.javax.xml.transform.stream);
importPackage(Packages.org.w3c.dom);
importPackage(Packages.com.alarmpoint.servicemanager.cm);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessage);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLToken);
importClass(Packages.com.alarmpoint.integrationagent.util.EventDeduplicator);
importClass(Packages.com.thoughtworks.xstream.XStream);
importClass(Packages.com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider);

var xStream = new XStream(new PureJavaReflectionProvider())

var INT_VERSION = "hpsmcm-3-0";

// Core Javascript files provided by the IA
load("integrationservices/" + INT_VERSION + "/lib/javascript/core/baseclass.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/core/logger.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/core/util.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/webservices/wsutil.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/webservices/soapfault.js");
load("integrationservices/" + INT_VERSION + "/lib/javascript/xmatters/xmattersws.js");

load("integrationservices/" + INT_VERSION + "/configuration.js");

var exceptionString = "JavaException: com.alarmpoint.servicemanager.cm.ServiceManagerException: ";

var notePrefix = "[xMatters] - "; // The Prefix which is added to the note comments e.g. "[xMatters] "
var noteType = "xMatters";        // The Activity Type of the history item"

var TOKEN_NAME_FORMATTED_EXCEPTION = "apiaformattedexception";  // Optional pre-formatted text for notification that will be sent to the user 

var LOCKED_RECORD_INDICATOR = "Requested resource is in use by"; // an indicator of Service Manager's "Record is locked" exception 

var smConnection = null;

var calloutAnnotatePass = null;

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by APClient.  The messages may be requests to perform
 * local activity, or they may be requests to make submissions to xMatters.
 * <p>
 * Any APXMLMessage object that this method returns will be sent to xMatters
 * via the Integration Service's outbound queues.
 */
function apia_input(apxml)
{
  var webService = new xMattersWS();
  log.info("Received an injection with tokens " + webService.convertTokensToString(apxml));
  
  // Forward enriched APXML to xMatters, subject to deduplication
	try
	{
    if (webService.submitAPXML(apxml, DELETE_EXISTING_EVENTS))
    {
        // If submitAPXML() actually sent the message and didn't deduplicate it, add an annotation to the IA logs.
        log.info(notePrefix + "Notification will be sent to [" + apxml.getValue("recipients") + "].");
    }
	}
  catch (ex)
  {
    log.error("An exception was encountered while sending APXML. Exception is: " + ex.toString());
    throw ex;
  } 
}

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by xMatters.  The messages may be responses to
 * previous submissions, or they may be requests to the Integration Service
 * that originate from xMatters
 * (e.g., via an ExternalServiceMessage/ExternalServiceRequest2).
 * <p>
 * Any APXMLMessage object that this method returns will be sent to xMatters
 * either via the Integration Service's outbound queues or directly as a
 * Web Service response, depending on the mechanism used to submit the APXML.
 */
function apia_response(apxml)
{
  var webService = new xMattersWS();
  log.info( "apia_response() received APXML from xMatters; content is: " + webService.convertTokensToString(apxml));

  var method = apxml.getMethod();
  
  log.debug("Method =[" + method + "]");
  
  var result = null;
      
  if (APXMLMessage.OK_METHOD.equalsIgnoreCase(method))
  {
    result = handleOK(apxml);
  }
  else if (APXMLMessage.ERROR_METHOD.equalsIgnoreCase(method))
  {
    result = handleERROR(apxml);
  }
  else if (APXMLMessage.SEND_METHOD.equalsIgnoreCase(method))
  {
    result = handleSend(apxml);
  }
  else if (APXMLMessage.REQUEST_METHOD.equalsIgnoreCase(method))
  {
    result = handleRequest(apxml);
  }
  else
  {
    throw new IllegalArgumentException("Unrecognized APXML method: " + method);
  }
  
  log.debug("Exit - apia_response");
  return result;
}


/**
 * OK responses are returned by xMatters for each APXML message that it accepts
 * from an Integration Service.
 * <p>
 * This implementation just logs the ack.
 */
function handleOK(apxml)
{
  var id = apxml.getTransactionId();
  log.debug("Received 'OK' response from xMatters for APXML message with ID: " + id);
}

/**
 * ERROR responses are returned by xMatters for each APXML message that it
 * is unable to accept from an Integration Service.
 * <p>
 * This implementation just logs the error.
 */
function handleERROR(apxml)
{
  var transaction = apxml.getTransactionId();
  var code = apxml.getValue(APXMLMessage.ERROR_CODE);
  log.error("Submission of transaction " + id + " failed with error code " + code);
}

/**
 * This implementation just logs query request, there shouldn't be any direct calls to this method
 */
function handleRequest(apxml)
{
  log.debug( "\tEnter - handleRequest");
  var result = handleSend(apxml);
  
  log.debug( "\tExit - handleRequest");
  return result;
}

/**
 * The main routine for handling Requests xMatters
 * <p>
 * Throw an exception that will be logged by the Integration Agent.
 */
function handleSend(apxml)
{
  log.debug( "\t\tEnter - handleSend");
  var result = null;
  
  // handle annotations coming from callout APS script 
  // they are formatted differently 
  // the presence of the message token means this message 
  // is a callout annotation. 
  var message = apxml.getValue("message");
  
  if ( message != null )
  {
    var changeId = apxml.getToken("change_id").getValue();
    result = handleCalloutAnnotate( changeId, message );
  }
  else
  {
    return handleApsResponses( apxml );
  }
  
  log.debug( "\t\tExit - handleSend");
  return result;
}
  
  
/**
 * The main routine for handling Requests from xMatters HP SM CM APS
 */  
function handleApsResponses( apxml )
{
  log.debug( "\t\t\tEnter - handleApsResponses");
  
  // Initialize xMattersWS interface building a reply to xMatters.
  var xmWebService = new xMattersWS();
  var responseToXM = xmWebService.initializeResponse(apxml);
  var responseTokens = getTokens(apxml);

  try
  {
    // deserialize the additional tokens
    log.debug( "\t\t\thandleApsResponses - Token keys from xMatters: " + responseTokens);

    var responseAction = responseTokens.get("responseAction");
    var annotation = responseTokens.get("annotation");

    // See if the APXML message is invalid, or if we don't have anything to do
    // in response to it.
    if (!checkAPXMLTokens(responseTokens))
    {
      // Nothing to do in response to this message, so just return OK
      responseToXM.setToken(xmWebService.TOKEN_NAME, "OK", APXMLToken.Type.STRING);
    }
    else
    {
      if (responseTokens.get("responseEvent") == "SUCCESSFUL_DELIVERY"
       || responseTokens.get("responseEvent") == "DELIVERY_FAILURE")
      {
        log.debug("Received delivery notification [" + responseTokens.get("responseEvent") + "]");
        // Create an annotation for the successful or failed delivery of a notification.
        // Set the responseAction token to Annotate to pass control to the Annotate handler.
        responseAction = "Annotate";
        annotation = responseTokens.get("responseEvent").toProperCase() + " to " + responseTokens.get("responder");
      }
 
      handleResponseAction(responseAction, annotation, responseTokens, responseToXM);

      responseToXM.setToken(xmWebService.TOKEN_NAME, "OK", APXMLToken.Type.STRING);
    }
  }
  catch (e)
  {
    var errMsg = "Caught exception processing [" + responseTokens.get("responseEvent") + "|" + responseAction + "]: " + e;
    log.error(errMsg);
    if ( errMsg.indexOf(LOCKED_RECORD_INDICATOR) > 0 ) 
    {  
      var changeId = responseTokens.get("change_id");
      var exceptionMessage = "Your response '" + responseTokens.get("responseEvent") + "|" + responseAction + "' was not processed as the HP SM record " + changeId +
         " is locked. To record your response, log into HP Service Manager and update the change manually. \n\nHP Service Manager returned the following error message:\n\n" + e;
      responseToXM.setToken(TOKEN_NAME_FORMATTED_EXCEPTION, exceptionMessage, APXMLToken.Type.STRING);   
      responseToXM.setToken("removeresponses", true);
      responseToXM.setToken("recordlocked", true);
    }            
    responseToXM.setToken(xmWebService.TOKEN_NAME_EXCEPTION, errMsg.replace(exceptionString, ""), APXMLToken.Type.STRING);
  }  
 
  log.debug( "Response about to be sent to xMatters: " + responseToXM );
  
  log.debug( "\t\t\tExit - handleApsResponses");
  return responseToXM;
  
}

/*
 * -----------------------------------------------------------------------------
 * handleResponseAction
 *
 * Dispatch the user response to the appropriate handler function. The caller
 * may have over-ridden the responseAction and annotation tokens from the APXML
 * message, but the other tokens will be untouched.
 * -----------------------------------------------------------------------------
 */
function handleResponseAction(responseAction, annotation, responseTokens, responseToXM)
{
   log.debug("\t\t\tEnter - handleResponseAction. Handling responseAction [" + responseAction + "] for changeID [" + responseTokens.get("changeId") + "]");

    var responder = responseTokens.get("responder");
    var responderId = responseTokens.get("responderId");
    var sendAnnotations = "true".equalsIgnoreCase(responseTokens.get("sendAnnotations"));
    var isRepliedFromXM = true;
    
    var smUser = responseTokens.get("sm_user");
    var smPass = responseTokens.get("sm_pass");

    if ( smUser == null || smUser == "" || smPass == null || smPass == ""|| smPass == "[]") 
    {
      logInfo("Credentials in apxml is missing sm_user or sm_pass which are both manditory will be using default credentials for HPSM_USER_NAME: " + HPSM_USER_NAME);
      smUser = HPSM_USER_NAME;
      // get the file handler for the encoded password
      smPass = new File(HPSM_PASSWORD_FILE);
    }
    
    // create the connection to the SM Script Object
    smConnection = ServiceManagerScriptObject( HPSM_URL, smUser, getPassword(smPass) );

    // extract message data
    var changeId = responseTokens.get("change_id");
    
    if (annotation != null)
    {
        annotation = notePrefix + annotation;
    }
    
    switch (String(responseAction).toLowerCase())
    {
        case "approve":
            log.debug("\t\t\t" + changeId + ": Approved by " + responder);
            smConnection.approveChange(changeId, noteType, notePrefix + "Approved by " + responder);
            break;
        case "deny":
            log.debug("\t\t\t" + changeId + ": Denied by " + responder);
            smConnection.denyChange(changeId, noteType, notePrefix + "Denied by " + responder);
            break;
        case "annotate":
            log.debug("\t\t\t" + changeId + ": annotation " + annotation + "]");
            smConnection.annotateChange(changeId, noteType, annotation);
            break;
        default:
            throw "Unknown response Action: " + responseAction;
            break;
    }

    log.debug("\t\t\tExit - handleResponseAction");
}


/*
 * checkAPXMLTokens
 *
 * Check that the received response tokens are valid and see if this message needs to be acted on. 
 * Throws an exception if message is invalid.
 * Returns true if the message needs to be processed.
 */
function checkAPXMLTokens(responseTokens)
{
    var result = true;

    switch (String(responseTokens.get("responseEvent")))
    {
        case "RECEIVED_RESPONSE":
        case "STANDALONE_ANNOTATION":
            if(responseTokens.get("responseAction") == null
                || responseTokens.get("responseAction") == ""
                || responseTokens.get("responseAction") == undefined)
                {
                throw "responseAction not set for a RECEIVED_RESPONSE responseEvent. responseAction required check xMatters scripting to ensure this token is set";
            }
            break;

        case "DEVICE_DELIVERED":
            // Log voice device delivery, but do not annotate to the MS, a callout annotation will be used instead.
            // So just return.
            log.debug(notePrefix + responseTokens.get("responseEvent") + " to " + responseTokens.get("responder"));
            result = false;
            break;

        case "SUCCESSFUL_DELIVERY":
        case "DELIVERY_FAILURE":
            if (!ANNOTATE_DELIVERY)
            {
                log.warn("DELIVERY_ANNOTATIONS are suppressed the following message will not be annotated to event: ["
                    + responseTokens.get("incidentId") + "] " + responseTokens.get("annotation"));
                result = false;
            }
            break;

        default:
            throw "Unknown responseEvent.  Expected: RECEIVED_RESPONSE||SUCCESSFUL_DELIVERY||DELIVERY_FAILURE||DEVICE_DELIVERED. Received: " 
            + responseTokens.get("responseEvent");
            break;
    }
   
    return result;
}

/**
 * handles annotations sent from the Callout APS script
 */
function handleCalloutAnnotate( incidentId, message )
{
  log.debug("Enter - handleCalloutAnnotate - incidentId [" + incidentId + "], messageText [" + message + "]");

  if(!ANNOTATE_DELIVERY)
  {
    log.warn("DELIVERY_ANNOTATIONS are suppressed the following message will not be annotated to event: [" + incidentId + "] " + message);
    return;
  }
  
  if (calloutAnnotatePass == null) 
  {
    calloutAnnotatePass = getPassword( new File(calloutAnnotatePassFile) );
    log.debug("calloutAnnotatePass: [" + calloutAnnotatePass + "]")
  }
  
  smConnection = ServiceManagerScriptObject( HPSM_URL, calloutAnnotateUser, calloutAnnotatePass );
  
  // incidentId is of the form C109876_Approval_7898 or C109876_Change_Logging 
  // so to get the changeId we need to get the stirng up to the _ 
  var changeId = getChangeId(incidentId); 

  log.debug("Starting call to management system for callout annotate for [" + changeId + "]");

  try 
  {
    smConnection.annotateChange( changeId, noteType, message );
  }
  catch (e)
  {
    logError("Caught exception processing annotation for [" + changeId + "] Exception:" + e);
  }
  
  log.debug("Exit - handleCalloutAnnotate");
}

// -----------------------------------------------------------------------------
// Utility functions
// -----------------------------------------------------------------------------

/**
 * Retrieves the additional tokens from the APXML
 * @param apxml XML message to process
 * @return additional tokens
 * @throws if the additional tokens are missing from the XML.
 */
function getTokens(apxml)
{
    if (apxml.getValue("additionaltokens") != null && !"".equalsIgnoreCase(apxml.getValue("additionaltokens")))
    {
        return xStream.fromXML(apxml.getValue("additionaltokens")); // LinkedHashMap
    }
    throw new Exception("additionaltokens is missing from apxml");
}

/**
 * Retrieves the password from the encryted password provided
 * @param encrypted password string
 */
function getPassword(encryptedPassword)
{ 
  var encryptionUtils = new EncryptionUtils(); 
  var password = encryptionUtils.decrypt(encryptedPassword); 

  return password;
}

/**
 * Converts incidentId that comes in the form C109876_Approval_7898 or C109876_Change_Logging into changeId
 * @param incidentId incidentId
 * @return changeId
 */
function getChangeId(incidentId) 
{
  var pos = incidentId.indexOf( "_" );
  var changeId; 
  if ( pos > 0 )
  {
    changeId = incidentId.substring( 0, pos );
  }
  else 
  {
    changeId = incidentId; 
  }
   
  return changeId;
}
 
var PREAMBLE = "Agent Client: HP SM CM - "
function logDebug(msg)
{
    ServiceAPI.getLogger().debug(PREAMBLE + msg);
} 

function logInfo(msg)
{
    ServiceAPI.getLogger().info(PREAMBLE + msg);
} 

function logError(msg)
{
    ServiceAPI.getLogger().error(PREAMBLE + msg);
} 
